(*
This file is part of the Kind verifier

* Copyright (c) 2007-2011 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** Base process module *)

(** 
@author Temesghen Kahsai

*)

open Types
open Flags
open Exceptions
open Channels
open Globals
open Mpi


(*Specific solver *)
let solver = Globals.my_solver
let toss x = ()  (* toss outputs *)
let refinement_pass = ref false (* set to true if doing another pass after refining *)
let newly_refined_vars = ref [] (* all vars refined this base k *)  
    
(* Defining some internal variables *)
let local_firsttime = ref true 
let final_position = 0 
let base_error = ref false 


(** Send/Recieve messages *)
let send_recieve_msg current_k prop_name =
  let _ = debug_proc BASE_PROC "Received message from Step Proc.\n" in
  let msg = receive step_proc any_tag comm_world in
    match msg with
	VALID_STEP kappa -> 
	  begin
	    if (kappa >= current_k) then
	      begin
		let _ = debug_proc BASE_PROC "EXTRA CHECKS ARE NEEDED" in
		let _ = debug_proc BASE_PROC ("K recivied = "^(string_of_int kappa) ^ "Current K = " ^(string_of_int current_k)^ " \n") in
		  ()
	      end
	    else 
	      begin
		let k_s = string_of_int kappa in
		let _ = Globals.base_time_stop := (wtime()) in
		let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
		let bt_string = string_of_float base_time in 
		  if !Flags.xml then (
		    print_xml(Kind_support.print_resultProp_xml 
				{result=VALID;
				 foundvars= None;
				 minstep=None;
				 maxstep=None;
				 induct_cex= None;
				 name="Specified property";
				 k=k_s;
				 time=bt_string});
		  ) else (
		    print_to_user_final ("\n\n+++   PROPERTY VALID with K = "^k_s^"    +++\n\n");
		    print_to_user_final("+++   WALL TIME (ms) :- "^bt_string^"   +++\n\n")); 
		  if !Flags.mpi_abort then
		    (mpi_abort();
		    )else( 
		      proc_size ==2 or (debug_proc BASE_PROC "Sent a msg to INV GEN"; isend STOP 2 9 comm_world;true);())
	      end;
	  end
      | STOP ->
	  begin
	    let k_s = "K" in
	    let _ = Globals.base_time_stop := (wtime()) in
	    let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	    let bt_string = string_of_float base_time in 
	      if !Flags.xml then (
		print_xml(Kind_support.print_resultProp_xml 
			    {result=UNKNOWN;
			     foundvars= None;
			     minstep=None;
			     maxstep=None;
			     induct_cex= None;
			     name="Specified property";
			     k=k_s;
			     time=bt_string});
	      ) else (
		print_to_user_final ("\n\n+++   UNKNOWN RESULT (maximum K iterations reached)  +++\n\n"));
	      if !Flags.mpi_abort then
		(mpi_abort();
		)else(
		  proc_size ==2 or (debug_proc BASE_PROC "Sent a msg to INV GEN";isend STOP 2 9 comm_world;true);())
	  end
      | STOP_INDUCT(induct_cex,global_k) ->
	  begin
	    let _ = Globals.base_time_stop := (wtime()) in
	    let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	    let bt_string = string_of_float base_time in 
	      if !Flags.xml then (
			print_xml(Kind_support.print_resultProp_xml 
				    {result=INDUCT_CEX;
				     foundvars= None;
				     minstep=None;
				     maxstep=None;
				     induct_cex= None;
				     name="Specified property";
				     k=global_k;
				     time=bt_string});
	      ) else (
		print_to_user_final("INDUCTIVE COUNTEREXAMPLE at K = "^ global_k ^"\n");
		print_to_user_final(induct_cex);
		print_to_user_final ("\n\n+++   UNKNOWN RESULT (maximum K iterations reached)  +++\n\n"));
	      if !Flags.mpi_abort then
		(mpi_abort();
		)else( 
		  proc_size ==2 or ( debug_proc BASE_PROC "Sent a msg to INV GEN";isend STOP 2 9 comm_world;true);())
	  end
      |  _ -> 
	   begin
	     debug_proc BASE_PROC "Recivied a wrong message";
	     Globals.base_time_stop := (wtime())
	   end


(** Solver response *)
let solver_response sol k first_time_n_values= 
  let k_s = string_of_int k in
  let k1 = k-1 in
  let out_base = solver#get#get_solver_output BASE_PROC sol.from_solver_ch in
    if (solver#get#result_is_unsat out_base) then
      begin (* base valid *)
	let _ = print_to_solver solver#get#safe_pop in
	let _ = debug_proc BASE_PROC "Valid" in
	let _ = print_to_solver solver#get#pop_string in
          (* re-assert the base defs for all newly refined vars *)
        let _ = 
	  List.iter (fun y ->
		       Kind_support.print_initialization_single sol.def_hash sol.startdepth sol.add y;
		       for i=sol.startdepth to k-1 do
			 Kind_support.persistent_step_single_assert sol.def_hash sol.startdepth 
			   sol.add BASE i y
		       done
		    ) !newly_refined_vars in
	  
	let _ = if !local_firsttime then
          begin
	    List.iter (fun x -> print_to_solver
			 (Lus_convert_yc.simple_command_string
			    (ASSERT(
			       F_PRED("P",[NUM(sol.add*(x+1))])
			     )))
		      ) (0::first_time_n_values)
          end
        else
          print_to_solver 
            (Lus_convert_yc.simple_command_string
               (ASSERT(
                  F_PRED("P",[NUM(sol.add*k)])
                ))) in
	  (*Incrementing K*)
        let _ = Kind_support.persistent_step_asserts_concrete sol.def_hash sol.startdepth sol.add k sol.from_checker_ch in
        let _ = refinement_pass := false in
        let _ = local_firsttime := false in
          true, k+1
    end (* base valid *)	  
  else if (solver#get#result_is_sat out_base) then
    begin (* base invalid *)
      let _ = debug_proc BASE_PROC "Invalid." in
      let basestep = if !local_firsttime then BASE_INIT else BASE in
      let foundvars = solver#get#get_cex BASE_PROC out_base print_to_solver sol.from_solver_ch in
      let _ = debug_proc BASE_PROC ("checking absmode "^(string_of_int (Kind_refine.get_num_abstract base_abstr))) in
      let _ = print_to_solver solver#get#safe_pop in
        if !abs_mode && Kind_refine.is_not_fully_refined base_abstr then
          begin
	    match Kind_refine.check_and_refine_abstraction sol.from_checker_ch 
	      sol.def_hash sol.startdepth sol.add basestep foundvars sol.pvars k with
                  CHECK_IS_VALID -> () (* continue *)
                | CHECK_IS_INVALID xs ->
			(* assert new vars *)
		    List.iter (fun y ->
				 (* assert just y up to base step *)
				 Kind_support.print_initialization_single sol.def_hash 
				       sol.startdepth sol.add y;
				 (* assert just y from sol.startdepth to k-1 *)
				 for i=sol.startdepth to k-1 do
				   Kind_support.persistent_step_single_assert sol.def_hash 
				     sol.startdepth sol.add basestep i y
				 done
			      ) xs;
		    (* note these new vars *)
		    newly_refined_vars :=  List.rev_append (!newly_refined_vars) xs;
		    refinement_pass := true;
          end
        else
          begin	       
	    let _ = debug_proc BASE_PROC ("Invalid at K ="^k_s^".\n") in   
	      if !checker_mode then
		toss(Kind_support.query_checker sol.from_checker_ch foundvars basestep k1);
	      let _ = Globals.base_time_stop := (wtime()) in
	      let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	      let bt_string = string_of_float base_time in 
		if !Flags.xml then (
		  print_xml (Kind_support.print_resultProp_xml 
			       {result=INVALID;
				foundvars= Some foundvars;
				minstep=Some 0;
				maxstep=Some k1;
				induct_cex= None;
				name="Specified property";
				k=k_s;
				time=bt_string});
		) else (
		  let _ =  print_to_user_final("FOUND COUNTEREXAMPLE at K ="^k_s^"\n\n") in
		      let _ = Kind_support.print_countermodel foundvars 0 k1 in
    	      		print_to_user_final("\n\n+++   WALL TIME (ms) :- "^bt_string^"   +++\n\n");
		    );
		if !Flags.mpi_abort then (
		  mpi_abort();
		) else (  
		  let _ = isend STOP step_proc 8 comm_world in
		    proc_size ==2 or (isend STOP 2 9 comm_world;true);())
	  end;
	false,k
    end (* base invalid *)
  else
    begin
      print_to_user_final ("Abort in "^k_s^" base due to incompleteness or error.\n");
      if (solver#get#result_is_error out_base) then
        print_to_user_final ((Str.matched_string out_base)^"\n");
      let _ = debug_proc BASE_PROC ("SOLVER OUTPUT: "^out_base) in
	Globals.base_time_stop := (wtime());
          base_error := true;
          false,k	
    end
      
      

(** Single property verification *)
let singleProp sol = 
  let _ = debug_proc BASE_PROC "SINGLE PROP." in
  (* Print initial case *)
  let _ = print_to_solver solver#get#push_string in
  let _ = Kind_support.print_initialization sol.def_hash sol.startdepth sol.add sol.from_checker_ch in
    
  (* For Abstraction/refinemnet *) 
  let _ = Kind_refine.set_num_abstract () in
  let _ = Kind_refine.set_not_fully_refined() in
  let k_inc = ref sol.startdepth in
  let do_bmc = ref true in
  let first_time_n_values = (Array.to_list (Array.init (!k_inc-1) (fun x -> x+1))) in

let rec iter_bmc do_bmc k = 
  let _ = debug_proc BASE_PROC "MAIN LOOP." in
  let k1 = k-1 in
  let k_s = string_of_int k in
 
    (* Check if recivied any new messages *)
    if iprobe step_proc any_tag comm_world != None then 
      send_recieve_msg k
    else
      begin
	let _ = debug_proc BASE_PROC "Not received message." in
	let _ = debug_proc BASE_PROC ("Checking k="^k_s) in
	if not !refinement_pass then
	  begin (* not refinement_pass *)
	    let _ = debug_proc BASE_PROC "not refinement_pass" in	
	    let _ =
	      if Lus_assertions.assertions_present() then
		begin
		  let _ = debug_proc BASE_PROC "Has assertion." in
		    print_to_solver (
		      Lus_convert_yc.simple_command_string
			(ASSERT(
			   F_PRED(solver#get#assertions,[NUM(sol.add*k1)]))));
		    if !local_firsttime then
		      for i= 0 to k1-1 do
			print_to_solver (
			  Lus_convert_yc.simple_command_string
                            (ASSERT(
			       F_PRED(solver#get#assertions,[NUM(sol.add*i)]))));
			debug_proc BASE_PROC "Has assertions, Done. ";
		      done
		end 
	    in	 
	    let _ = print_to_solver solver#get#push_string in
	      
            (* Assert the properties here*)
            let impl_premise = F_EQ(POSITION_VAR(sol.nstart),NUM(sol.add*k1),L_INT) in
            let impl_result = 
	      if !local_firsttime && k > 1 then
		begin
		  List.fold_left (fun acc x -> 
				    F_AND(F_PRED("P",[NUM(sol.add*x)]), acc)) (F_PRED ("P",[NUM(0)])) 
                    first_time_n_values
		end
              else
		F_PRED("P",[NUM(final_position)])
            in
            let _ = print_to_solver(Lus_convert_yc.simple_command_string (
				      ASSERT(
					F_NOT(F_IMPL(impl_premise,impl_result))))) in
            let _ = print_to_solver solver#get#safe_push (* PUSH 2a *)  in
              print_to_solver ( Lus_convert_yc.simple_command_string (QUERY(F_TRUE)))
	  end (* not refinement_pass *)
	else
	    begin
              let _ = print_to_solver solver#get#safe_push in
		print_to_solver (Lus_convert_yc.simple_command_string (QUERY(F_TRUE)))
	    end;
	  print_to_solver solver#get#done_string;
	  let do_bmc_next, new_k = solver_response sol k first_time_n_values in
	    iter_bmc do_bmc_next new_k 
      end;
in
  iter_bmc true !k_inc


(** Initialization *)
let init_baseProc filename =
  let defdoc,maxdepth,def_hash,no_stateful,pvars,prop_list = Defgen.start filename in
  let add = Kind_support.get_add() in
  let nstart = solver#get#step_base_string in
  let startdepth = maxdepth + 1 in
  let from_checker_ch, from_solver_ch = Kind_support.setup_solver1() in
 
    if (!termination_check || !loopfree_mode) && not !abs_mode then (
      static_path_compression.(base_abstr) <- true;
      static_path_compression.(step_abstr) <- true);
    (* Needed for refinement*)
    if !checker_mode then print_to_checker solver#get#checker_setup_string;
    (* Print definitions *)
    let _ = Kind_support.print_defs_to_solver1 from_solver_ch from_checker_ch (defdoc^"\n") in	  
	  { 
	    pvars = pvars;
	    add = add;
	    startdepth = startdepth;
	    nstart = nstart;
	    nvar = "";
	    from_solver_ch = from_solver_ch; 
	    from_solver2_ch = dummy;
	    from_checker_ch = from_checker_ch;
	    cur_depth = maxdepth;
	    def_hash = def_hash;
	    no_stateful = no_stateful;
	    prop_list = prop_list;
	    defdoc = "";
	  }
	    
	    
(** Starting point for base proc*)
let main filename = 
  let sol = init_baseProc filename in
    (* If only single property do single BMC, otherwise do multi BMC (i.e. incremental verification)*)
    if List.length sol.prop_list == 1 && !Flags.multi then
       (debug_proc BASE_PROC "SINGLE";
       singleProp sol
      ) else (
	if  List.length sol.prop_list == 1 then
	  (debug_proc BASE_PROC "SINGLE PROPERTY";
	   singleProp sol
	  ) else (
	    debug_proc BASE_PROC "MULTIPLE PROPERTIES";
	    Multi_bmc.main sol)
      )
