(*
This file is part of the Kind verifier

* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(* PKIND extra checks k-induction  *)


(** A module for rechecking the results of certain properties. *)

(** 
@author Temesghen Kahsai

*)

open Types
open Flags
open Exceptions
open Channels
open Globals
open Mpi
open Util

let solver = Globals.my_solver 

(** Function to do extra checks during incremental multiple generation *)
(* let extra_unsat_checks defdoc pvars maxdepth sol.def_hash unsat_prop_list inv_generated prop_as_inv k = *)

let induct_check sol unsat_prop_list inv_generated prop_as_inv k = 
  let real_unsat = ref false in
  let k1 = !k-1 in
  let k_s = string_of_int !k in
  let firststep = ref true in
  let step_error = ref false in
  let switch_to_static_compression = ref false in
  let perform_step_check = ref true in
  let initialize_step = ref true in
  let previous_static_compression = ref static_path_compression.(step_abstr) in 
  let new_unsat_list = ref [] in 
  let from_checker_ch, from_solver5_ch = Kind_support.setup_solver_induct_checker () in

   (*Creating lists*)
  let var_prop =  List.map (fun x -> (Tables.get_varinfo_name x)) sol.pvars in
  let unsat_var_prop_list  = ref [] in
  let sat_var_prop_list  = ref [] in 
  let unsat_id_prop_list  = ref [] in
  let sat_id_prop_list  = ref [] in
  let prop_to_be_checked = ref [] in
    

  (* Initial list of properties*)
  unsat_var_prop_list := unsat_prop_list; 
  sat_var_prop_list := unsat_prop_list; 
  unsat_id_prop_list := List.map (fun x -> Tables.varid_lookup x) unsat_prop_list;
  sat_id_prop_list := List.map (fun x -> Tables.varid_lookup x) unsat_prop_list;
  prop_to_be_checked := unsat_prop_list;

   List.iter (fun x -> debug_proc INDUCT_PROC(string_of_int x)) !unsat_id_prop_list;

   debug_proc INDUCT_PROC ("MULTI EXTRA CHECKS: Performing extra checks on UNSAT props for K="^k_s);

 (* print definitions *)
  Kind_support.print_defs_to_solver_induct_checker from_solver5_ch (sol.defdoc^"\n");
 
  let all_checked = ref true in
    
    while (!all_checked) do
      if !initialize_step || not !previous_static_compression then
   begin (* setup step loop *)
     initialize_step := false;
       debug_proc INDUCT_PROC "MULTI EXTRA CHECKS: setup step loop - Normal initialize step.";
       if !loopfree_mode && not static_path_compression.(step_abstr) then
         begin
           try
             if !switch_to_static_compression then
               begin
                 switch_to_static_compression := false;
                 raise AllStatefulVarsRefined
               end;
	     
             let header_string = solver#get#push_string^(Lus_convert_yc.yc_state_vars_string_refined step_abstr)^"\n" in
               print_to_checker solver#get#push_string;
	       debug_proc INDUCT_PROC ";4";
               print_to_solver5 header_string;
             Kind_compression.abstracted_compression_constraint sol.startdepth sol.add print_to_solver5 0 (k1-sol.startdepth) sol.nvar ;
           with NoStatefulVars -> 
	     debug_proc INDUCT_PROC ";4e";
             print_to_solver5 solver#get#push_string(***** PUSH *****) 
             | AllStatefulVarsRefined ->
                 static_path_compression.(step_abstr) <- true;
                 Kind_compression.general_compression_constraint sol.startdepth sol.add print_to_solver5 0 (k1-sol.startdepth) sol.nvar;
                 previous_static_compression := true;
		 debug_proc INDUCT_PROC ";4f";
                 print_to_solver5 solver#get#push_string(***** PUSH *****)
         end
       else
         begin
           print_to_solver5 solver#get#push_string;(***** PUSH *****)
	   debug_proc INDUCT_PROC ";4b";
           previous_static_compression := true
         end;
       
       
       for i=0 to (!k-1) do
         if !var_defs_mode then
	   Kind_support.def_assert_both5 sol.def_hash "DEF" [PLUS(POSITION_VAR(sol.nvar),NUM(i*sol.add))] !k from_checker_ch
         else
           begin
             print_to_solver5
               (Lus_convert_yc.simple_command_string
		  (ASSERT(
                     F_PRED("DEF",[PLUS(POSITION_VAR(sol.nvar),NUM(sol.add*i))])
                   )))
           end
       done;

       (*Assert isol.nvariants discovered if any *)
       if not (List.length inv_generated == 0) then
	 (debug_proc INDUCT_PROC  "MULTY EXTRA CHECK: DISCOVERED INVARIANTS";
	  List.iter (fun x -> Kind_support.print_defs_to_solver_induct_checker from_solver5_ch (x^"\n")) inv_generated;
	  for i=0 to (!k-1) do   
            print_to_solver5 (
                  Lus_convert_yc.simple_command_string
                    (ASSERT(F_PRED(solver#get#assertions_inv,[PLUS(POSITION_VAR(sol.nvar),NUM(sol.add*i))])))); 
	  done; 
	 );
       
       (*Assert proven properties if any *)
       if not (List.length prop_as_inv == 0) then
	 (debug_proc INDUCT_PROC "MULTY EXTRA CHECK: PROPERTY AS INVARIANTS";
	   List.iter( fun x -> print_to_solver5("(assert ("^ x ^ " (+ _n 0)))\n")) prop_as_inv;
	   for j=1 to (!k-1) do
	    List.iter( fun x -> print_to_solver5("(assert ("^ x ^ " (- _n "^string_of_int(j)^")))\n")) prop_as_inv
           done;
	 );
       
       for i=1 to (!k-1) do
	   List.iter( fun x -> print_to_solver5("(assert ("^ x ^ " (- _n "^string_of_int(i)^" )))\n")) !prop_to_be_checked
       done;
       if (List.length !prop_to_be_checked ==1) then
	 begin
	   print_to_solver5("(assert ( not ");
	   List.iter(fun x -> print_to_solver5("("^x^" _n)")) !prop_to_be_checked;
	   print_to_solver5("))"); 
	   print_to_solver5("\n"); 
	   end
       else
	 begin
	   print_to_solver5("(assert ( not ( and");
	   List.iter(fun x -> print_to_solver5("("^x^" _n)")) !prop_to_be_checked;
	   print_to_solver5(")))"); 
	   print_to_solver5("\n"); 
	 end;
     end (* setup step loop *)
 else
  begin (* alt setup step loop *)
 
    debug_proc INDUCT_PROC "MULTI EXTRA CHECKS: setup step loop - no initialiaze";
   print_to_checker solver#get#push_string;
     
   for i=(!k-sol.startdepth) to (!k-1) do
     if !var_defs_mode then
       Kind_support.def_assert_both5 sol.def_hash "DEF" [PLUS(POSITION_VAR(sol.nvar),NUM(i*sol.add))] !k from_checker_ch
     else
       begin
	 print_to_checker(Lus_convert_yc.simple_command_string
			    (ASSERT(F_PRED("DEF",[PLUS(POSITION_VAR(sol.nvar),NUM(sol.add*i))]))))
       end
   done;
   if !k_step_size > 1 then
     begin
       (* fill in for k_step_size > 1 *)
       let startpoint,endpoint = 
         if !firststep then 0,(!k-sol.startdepth)-1
         else
           !k - !k_step_size - sol.startdepth + 1, (!k-sol.startdepth)-1 in 
       let defname = "DEF" in
         for i=startpoint to endpoint do
           if !var_defs_mode then
	     Kind_support.def_assert_both5 sol.def_hash defname [PLUS(POSITION_VAR(sol.nvar),NUM(i*sol.add))] !k from_checker_ch
           else
             begin
	       print_to_checker (Lus_convert_yc.simple_command_string
	               (ASSERT(F_PRED(defname,[PLUS(POSITION_VAR(sol.nvar),NUM(sol.add*i))]))))
             end
         done
     end
  end;
 
let assert_base = (F_EQ(POSITION_VAR(sol.nstart),NUM(sol.add*k1),L_INT)) in
print_to_solver5 (Lus_convert_yc.simple_command_string (ASSERT assert_base));
print_to_solver5 solver#get#safe_push; (* SAFE PUSH 1 *)
print_to_solver5 (Lus_convert_yc.simple_command_string(QUERY(F_TRUE)));
print_to_solver5 solver#get#done_string;


let out_step_extra = solver#get#get_solver_output INDUCT_PROC from_solver5_ch in
  if (solver#get#result_is_unsat out_step_extra) then
    begin (* step valid *)  
      print_to_solver5 solver#get#safe_pop; (* SAFE PUSH 1 *)
      if !check_compression && !loopfree_mode && not static_path_compression.(step_abstr) then
          begin
            debug_proc INDUCT_PROC("MULTI EXTRA CHECKS: Re-checking with full compression.\n");
            switch_to_static_compression := true;
            print_to_solver5 solver#get#pop_string;
            (* sol.add the static def here *)
            initialize_step := true;
            perform_step_check := true 
          end
       else 
        begin
	  debug_proc INDUCT_PROC("\nMULTI EXTRA CHECKS: All property are valid -> [");
	  List.iter(fun x -> debug_proc INDUCT_PROC(x^" ")) !prop_to_be_checked;
	  debug_proc INDUCT_PROC("].\n");
	  all_checked := false;
	  real_unsat :=true;
	  new_unsat_list := !prop_to_be_checked;
        end 
    end (* step valid *)

else if (solver#get#result_is_sat out_step_extra || solver#get#result_is_unknown out_step_extra) then
begin (* step invalid *)
  if solver#get#result_is_unknown out_step_extra then
    debug_proc INDUCT_PROC ("MULTI EXTRA CHECKS: Unknown in "^k_s^" step.\n")
  else
    debug_proc INDUCT_PROC ("MULTI EXTRA CHECKS: step is Invalid at k ="^k_s^".\n");
   let foundvars = solver#get#get_cex INDUCT_PROC out_step_extra print_to_solver5 from_solver5_ch in
   let simulation_value_hash = solver#get#get_simulation_cex INDUCT_PROC out_step_extra print_to_solver5 from_solver5_ch  in
   let value_n = (Hashtbl.find simulation_value_hash ("_n")) in
   
    debug_proc INDUCT_PROC ("MULTI EXTRA CHECK: Updating list of properties.\n");
     
    unsat_id_prop_list := 
      List.filter (fun id -> Multi_prop_util.is_false 
		     (Hashtbl.find foundvars (id,int_of_string(value_n)))) !unsat_id_prop_list;

     sat_id_prop_list := 
      List.filter (fun id -> Multi_prop_util.is_true 
		     (Hashtbl.find foundvars (id,int_of_string(value_n)))) !sat_id_prop_list;
   
    unsat_var_prop_list := List.map (fun x -> (Tables.get_varinfo_name x)) !unsat_id_prop_list;
    
    sat_var_prop_list :=  List.map (fun x -> (Tables.get_varinfo_name x)) !sat_id_prop_list;
     
    debug_proc INDUCT_PROC("\n MULTI EXTRA CHECK: SAT Properties at K = "^k_s^" \n"); 
    List.iter(fun x -> debug_proc INDUCT_PROC(x^" ")) !unsat_var_prop_list;
    debug_proc INDUCT_PROC("\n MULTI EXTRA CHECK: UNSAT Properties at K = "^k_s^" \n"); 
    List.iter(fun x -> debug_proc INDUCT_PROC(x^" ")) !sat_var_prop_list;
    debug_proc INDUCT_PROC("\n");
    prop_to_be_checked := !sat_var_prop_list;
    if (List.length !prop_to_be_checked == 0) then
    ( debug_proc INDUCT_PROC("MULTI EXTRA CHECK: All SAT, return to STEP PROC\n");
      all_checked := false;
    ) else (
     print_to_solver5 solver#get#pop_string;
     initialize_step := true;
    debug_proc INDUCT_PROC ("MULTI EXTRA CHECK: Done Updating list of properties.\n");
    real_unsat := false;
    );
  end (* step invalid *)
else
  begin
    (* STEP check = ERROR *)
    debug_proc INDUCT_PROC ("MULTI EXTRA CHECKS: Abort in "^k_s^" step.\n");
    if (solver#get#result_is_error out_step_extra) then
      print_to_user_final ((Str.matched_string out_step_extra)^"\n");
      debug_proc INDUCT_PROC ("MULTI EXTRA CHECKS : "^out_step_extra);
      step_error := true
  end;

(*Resetting the current logical context *)
done;
print_to_solver5("(reset)");
!real_unsat,
!new_unsat_list

 
