(*
This file is part of the Kind verifier

* Copyright (c) 2007-2011 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(** Base process module for multiple properties *)

(** 
@author Temesghen Kahsai

*)

open Types
open Flags
open Exceptions
open Channels
open Globals
open Mpi
open ExtString


let solver = Globals.my_solver
let toss x = () (* toss outputs *)

let main sol = 
  let _ = debug_proc BASE_PROC "MULTI: MAIN LOOP." in
  let k_inc = ref sol.startdepth in
  let local_firsttime = ref true in 
    
  (* List of properties *)
  let position = POSITION_VAR (solver#get#position_var1) in 
  let parsed_multiple_prop  =  Lus_convert_yc.yc_multi_property_header_list position sol.prop_list in
  let var_prop = List.map (fun x -> String.strip ~chars:" " (Multi_prop_util.mk_name_property x)) parsed_multiple_prop in
  let named_prop = Multi_prop_util.create_new_named_list var_prop in
    
  let unsolved_prop_list = ref [] in
  let unsat_id_prop_list  = ref [] in
  let sat_id_prop_list  = ref [] in
  let unsat_var_prop_list  = ref [] in
  let sat_var_prop_list  = ref [] in 
  let recheck_sat_list = ref [] in
    
  (*Collection of valid/invalid properties*)
  let valid_properties = ref [] in
  let invalid_properties = ref [] in
    
  let _ = print_to_user_mpkind("\n\nThere are "^string_of_int(List.length named_prop)^" property(ies) to be checked.\n") in
  let _ = print_to_user_mpkind("PROPERTIES TO BE CHECKED: [") in
  let _ = List.iter(fun x ->  (print_to_user_mpkind(Multi_prop_util.get_real_prop_name (x)^" "))) named_prop in
  let _ = print_to_user_mpkind("]\n") in
    
    
  (* Initial list of properties*)
  let _ = unsolved_prop_list := var_prop in
  let _ = unsat_var_prop_list := var_prop in 
  let _ = sat_var_prop_list := var_prop in
  let _ = unsat_id_prop_list := sol.pvars in
  let _ = sat_id_prop_list := sol.pvars in
    
  let _ = print_to_solver solver#get#push_string in
  let _ = Kind_support.print_initialization sol.def_hash sol.startdepth sol.add sol.from_checker_ch in

  let base_error = ref false in
  let local_base_stop = ref false in

  (* For Abstraction/refinemenet *) 
  let _ = Kind_refine.set_num_abstract () in
  let _ = Kind_refine.set_not_fully_refined() in
  let refinement_pass = ref false in (* set to true if doing another pass after refining *)
  let newly_refined_vars = ref [] in (* all vars refined this base k *) 
    
 while not !local_base_stop do 
   if iprobe step_proc any_tag comm_world = None then
     begin
       let _ = debug_proc BASE_PROC "MULTI: Not received message." in  
       let k = !k_inc in
       let k1 = k-1 in
       let k_s = string_of_int k in
	 if not !refinement_pass then
	   begin
	     let _ = debug_proc BASE_PROC "MULTI: Not refinement_pass" in
	       (* handle assertions *)	
	     let _ = 
	       if Lus_assertions.assertions_present() then
		 begin (* handle assertions *)
		   let _ = debug_proc BASE_PROC "MULTI: Printing assertion." in
		     print_to_solver (
		       Lus_convert_yc.simple_command_string
			 (ASSERT(
			    F_PRED(solver#get#assertions,[NUM(sol.add*k1)]))));
		     if !local_firsttime then
		       for i= 0 to k1-1 do
			 print_to_solver (
			   Lus_convert_yc.simple_command_string
                             (ASSERT(
			      F_PRED(solver#get#assertions,[NUM(sol.add*i)]))));
			 debug_proc BASE_PROC "MULTI: Done printing assertion ";
		       done
		 end
	     in
	       if !local_firsttime && k > 1 then
		 begin
		   print_to_solver solver#get#push_string; (***** PUSH *****)    
		   print_to_solver("(assert (not (=> (= _base (- 0 "^string_of_int(k1)^"))");
		 if (List.length !unsolved_prop_list == 1) then
		   (
		     print_to_solver("(and ");
		     List.iter(fun x -> print_to_solver("("^x^" ")) !unsolved_prop_list;
		     print_to_solver("(- 0 "^string_of_int(k1)^"))");
		     List.iter(fun x -> print_to_solver("("^x^" ")) !unsolved_prop_list;
		     print_to_solver("0)))))\n");
		   ) else (
		     print_to_solver("(and ");
		     List.iter(fun x -> print_to_solver("("^x^"(- 0 "^string_of_int(k1)^"))")) !unsolved_prop_list;
		     List.iter(fun x -> print_to_solver("("^x^" 0)")) !unsolved_prop_list;
		     print_to_solver("))))\n");)
		 end
	       else
		 begin
		   let _ = print_to_solver solver#get#push_string in
		   let _ = print_to_solver("(assert (not ( => (= _base (- 0 "^string_of_int(k1)^"))") in
		 if (List.length !unsolved_prop_list == 1) then (
		   List.iter(fun x -> print_to_solver("("^x^" ")) !unsolved_prop_list;
		   print_to_solver("0))))\n");
		 ) else (
		   print_to_solver("(and ");
		   List.iter(fun x -> print_to_solver("("^x^" 0)")) !unsolved_prop_list;
		   print_to_solver("))))\n");)
		 end
     end (* not refinement_pass *)
   else
     begin
       let _ = print_to_solver solver#get#safe_push in
       let _ = print_to_solver (Lus_convert_yc.simple_command_string (QUERY(F_TRUE))) in ()
     end;	 
   let _ = print_to_solver ( Lus_convert_yc.simple_command_string (QUERY(F_TRUE))) in
   let _ = print_to_solver solver#get#done_string in

   let out_base = solver#get#get_solver_output BASE_PROC sol.from_solver_ch in
     if (solver#get#result_is_unsat out_base) then
       begin (* base valid *)
	 let _ = debug_proc BASE_PROC "MULTI: Valid base case" in
	 let _ = print_to_solver solver#get#pop_string in
	   (* re-assert the base defs for all newly refined vars *)
         let _ = 
	   List.iter (fun y ->
			Kind_support.print_initialization_single sol.def_hash sol.startdepth sol.add y;
			for i=sol.startdepth to k-1 do
			  Kind_support.persistent_step_single_assert sol.def_hash sol.startdepth 
			    sol.add BASE i y
			done
		     ) !newly_refined_vars in
	 let _ = 
	   for i=1 to k1+1 do
	     List.iter(fun x -> print_to_solver("(assert ("^x^" (- 0 "^string_of_int(i)^")))\n")) !unsolved_prop_list;
	   done in	
	 let _ = Kind_support.persistent_step_asserts_concrete sol.def_hash sol.startdepth sol.add k sol.from_checker_ch in 
	 let _ = local_firsttime := false in
	   incr k_inc
       end (* base valid *)
     else if (solver#get#result_is_sat out_base) then
       begin (* base invalid *)
	 let _ = debug_proc BASE_PROC "MULTI: Invalid base case" in 
	 let foundvars = solver#get#get_cex BASE_PROC out_base print_to_solver sol.from_solver_ch in
	 let _ = debug_proc BASE_PROC "MULTI: Updating list of properties" in
	 let basestep = if !local_firsttime then BASE_INIT else BASE in
	 let _ = 
	   if !abs_mode && Kind_refine.is_not_fully_refined base_abstr then
             begin
	       match Kind_refine.check_and_refine_abstraction sol.from_checker_ch 
		 sol.def_hash sol.startdepth sol.add basestep foundvars sol.pvars k with
                     CHECK_IS_VALID -> () (* continue *)
                   | CHECK_IS_INVALID xs ->
			(* assert new vars *)
		       List.iter (fun y ->
				    (* assert just y up to base step *)
				    Kind_support.print_initialization_single sol.def_hash 
				      sol.startdepth sol.add y;
				    (* assert just y from sol.startdepth to k-1 *)
				    for i=sol.startdepth to k-1 do
				      Kind_support.persistent_step_single_assert sol.def_hash 
					sol.startdepth sol.add basestep i y
				    done
			      ) xs;
		       (* note these new vars *)
		       newly_refined_vars :=  List.rev_append (!newly_refined_vars) xs;
		       refinement_pass := true
             end 
	 in

	 (* We classify the properties in SAT and UNSAT from the value ID that we get from the counterexample *)
	 let _  = sat_id_prop_list := 
	   List.filter (fun id -> Multi_prop_util.is_false 
			  (Hashtbl.find foundvars (id,0))) !sat_id_prop_list in
	 let _ = unsat_id_prop_list := 
	   List.filter (fun id -> Multi_prop_util.is_true 
			  (Hashtbl.find foundvars (id,0))) !unsat_id_prop_list in
	       
	     (*We update the VAR of the properties and we also update the list of unsolved properties *)
	 let _ = unsat_var_prop_list := List.map (fun x -> (Tables.get_varinfo_name x)) !unsat_id_prop_list in
	 let _ = sat_var_prop_list :=  List.map (fun x -> (Tables.get_varinfo_name x)) !sat_id_prop_list in
	 let _ = unsolved_prop_list := List.filter(fun x ->  not (List.mem x !sat_var_prop_list)) !unsolved_prop_list in
	   
	 (* Once we update the unsolved propo list, we update the sat and the unsat list*)
	 let _ = sat_id_prop_list := List.map (fun x -> Tables.varid_lookup x) !unsolved_prop_list in
	 let _ = unsat_id_prop_list := List.map (fun x -> Tables.varid_lookup x) !unsolved_prop_list in
	   
	 (* Update the real name of the properties.*)
	 let _ = invalid_properties := List.append !invalid_properties (List.map (fun x -> Multi_prop_util.get_real_prop_name x) !sat_var_prop_list) in
	 let _ = debug_proc BASE_PROC "MULTI: DONE Updating list of properties " in
	   Globals.base_time_stop := (wtime());
	   let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	   let bt_string = string_of_float base_time in 
	     if !Flags.xml then (
	       let prop_names =  Multi_prop_util.mk_prop_names !sat_var_prop_list in
		 print_xml (Kind_support.print_resultProp_xml 
			      {result=INVALID;
			       foundvars= Some foundvars;
			       minstep=Some 0;
			       maxstep=Some k1;
			       induct_cex= None;
			       name=prop_names;
			       k=k_s;
			       time=bt_string});
	     ) else (
	       print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
	       print_to_user_mpkind("PROPERTY [");
	       List.iter (fun x -> print_to_user_mpkind(Multi_prop_util.get_real_prop_name (x)^" ")) !sat_var_prop_list;
	       print_to_user_mpkind ("] is Invalid. ||  K = "^k_s^ " || Time = "^bt_string^" || COUNTEREXAMPLE:\n");
	       Kind_support.print_countermodel foundvars 0 k1; 
	       print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n"););
	     
	     if (List.length !unsolved_prop_list == 0) then
	       begin
		 Multi_prop_util.print_summary !valid_properties !invalid_properties;
		 if !Flags.mpi_abort then
		   (if !Flags.xml then (
			   print_xml ("</Results>");
			 );
		     mpi_abort();)
		 else
		   ( 
		     isend STOP step_proc 8 comm_world;
		     proc_size ==2 or (debug_proc BASE_PROC "Sent a msg to INV GEN";isend STOP 2 9 comm_world;true);
		     local_base_stop := true )
	       end
	     else
	       begin
		 let _ = debug_proc BASE_PROC "MULTI: Sending invalid prop list -> [" in
		   List.iter(fun x -> debug_proc BASE_PROC (x^" ")) !sat_var_prop_list;
		   debug_proc BASE_PROC "] to INDUCTIVE PROC.\n";
		   let _ = debug_proc BASE_PROC "Sending invalid property " in
		   let msg_to_ind_proc = M_STEP_INVALID(!sat_var_prop_list) in
		     isend msg_to_ind_proc step_proc 8 comm_world; 	      
		     let _ = print_to_solver solver#get#pop_string in
		       local_firsttime := false 
	       end;
	   end (* base invalid *)
	 else
	   begin
             if (solver#get#result_is_error out_base) then
               print_to_user_final ((Str.matched_string out_base)^"\n");
             let _ = debug_proc BASE_PROC ("SOLVER OUTPUT: "^out_base) in
	       Globals.base_time_stop := (wtime());
	       base_error := true;
               local_base_stop := true
	   end
     end
   else
(**********  Recieved a message ***********)
begin
  let _ = debug_proc BASE_PROC "MULTI: Received message from Step Proc." in
  let list_of_valid_proc = receive step_proc any_tag comm_world in
    match list_of_valid_proc with
	M_STEP_VALID(valid_prop_ind,new_k) -> 
	  begin
	    if (new_k >= !k_inc) then 
	      begin 
		let _ = debug_proc BASE_PROC "MULTI: Extra checks are needed " in
		let _ = valid_properties := List.append !valid_properties (List.map (fun x -> Multi_prop_util.get_real_prop_name x) valid_prop_ind) in
		let _ = unsolved_prop_list := List.filter(fun x -> not (List.mem x valid_prop_ind)) !unsolved_prop_list in

	     (* Get rid off the property which are already discovered to be invalid *)
		let _ = valid_properties := List.filter(fun x -> not (List.mem x !invalid_properties)) !valid_properties in
		let _ = recheck_sat_list :=  List.filter(fun x -> not (List.mem x !sat_var_prop_list)) valid_prop_ind in
		  
		(* Recheck the properties that are "tentatively" valid.*)
		let real_unsat, new_unsat_list =  
		  Extra_checks.induct_check sol !recheck_sat_list [] [] (ref new_k) in 
		  if (real_unsat) then
		    begin
		      debug_proc BASE_PROC ("Confirmed that properties are valid.");	
		      Globals.base_time_stop := (wtime());
		      let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
		      let bt_string = string_of_float base_time in 
		      let k_s = string_of_int new_k in
			if !Flags.xml then (
			  let prop_names =  Multi_prop_util.mk_string !valid_properties in
			    print_xml (Kind_support.print_resultProp_xml 
					 {result=VALID;
					  foundvars= None;
					  minstep=None;
					  maxstep=None;
					  induct_cex= None;
					  name=prop_names;
					  k=k_s;
					  time=bt_string});
			) else (
	      		  print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
			  print_to_user_mpkind("VALID PROPERTY: [");
			  List.iter (fun x -> print_to_user_mpkind(x^" ")) !valid_properties;
			  print_to_user_mpkind ("]. || K = "^k_s^ " || Time = "^bt_string^"\n");
			  print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n"););
			  
			if (List.length !unsolved_prop_list == 0) then
			  begin
			    Multi_prop_util.print_summary !valid_properties !invalid_properties;
			    if !Flags.mpi_abort then
			      (if !Flags.xml then (
				 print_xml ("</Results>");
			       );
			       mpi_abort();)
			    else
			      ( 
				proc_size ==2 or (debug_proc BASE_PROC "MULTI: Sent a msg to INV GEN";isend STOP 2 9 comm_world;true);
				Globals.base_stop := true;
			      )
			    end;
		    end
		  else
		    begin
		      debug_proc BASE_PROC (" Uncofirmed that properties are valid ");
		    end; 
	      end
	    else
	      begin
		valid_properties := List.append !valid_properties (List.map (fun x -> Multi_prop_util.get_real_prop_name x) valid_prop_ind);
		unsolved_prop_list := List.filter(fun x -> not (List.mem x valid_prop_ind)) !unsolved_prop_list;		
		Globals.base_time_stop := (wtime());
		let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
		let bt_string = string_of_float base_time in 
		let k_s = string_of_int new_k in 
		  if !Flags.xml then (
		    let prop_names =  Multi_prop_util.mk_prop_names valid_prop_ind in
		      print_xml (Kind_support.print_resultProp_xml 
				   {result=VALID;
				    foundvars=None;
				    minstep=None;
				    maxstep=None;
				    induct_cex= None;
				    name=prop_names;
				    k=k_s;
				    time=bt_string});
		  ) else (
	      	    print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
		    print_to_user_mpkind("VALID PROPERTY: [");
		    List.iter (fun x -> print_to_user_mpkind(Multi_prop_util.get_real_prop_name (x)^" ")) valid_prop_ind;
		    print_to_user_mpkind ("]. || K = "^k_s^ " || Time = "^bt_string^"\n");
	            print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n"););
		  
		  if (List.length !unsolved_prop_list == 0) then
		    begin
		      Multi_prop_util.print_summary !valid_properties !invalid_properties;
		      if !Flags.mpi_abort then
			(if !Flags.xml then (
			   print_xml ("</Results>");
			 );
			 mpi_abort();)
		      else
			( 
			  proc_size ==2 or (debug_proc BASE_PROC "MULTI: Sent a msg to INV GEN";isend STOP 2 9 comm_world;true);
			  local_base_stop := true;
			)
		    end;
	      end;  
	  end;
      | STOP_INDUCT(induct_cex,global_k) ->
	  begin
	    Globals.base_time_stop := (wtime());
	    let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	    let bt_string = string_of_float base_time in 
	      if !Flags.xml then (
		print_xml (Kind_support.print_resultProp_xml 
			     {result=UNKNOWN;
			      foundvars=None;
			      minstep=None;
			      maxstep=None;
			      induct_cex= Some induct_cex;
			      name="INDUCTIVE CEX";
			      k=global_k;
			      time=bt_string});
	      ) else (
		print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
		print_to_user_mpkind ("UNKNOWN RESULT (maximum K iterations reached) \n\n");
		print_to_user_mpkind("INDUCTIVE COUNTEREXAMPLE at K = "^ global_k ^"\n");
		print_to_user_mpkind(induct_cex);
		print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n"););
		
		if !Flags.mpi_abort then
		  (if !Flags.xml then (
		     print_xml ("</Results>");
		   );
		   mpi_abort();)
		else
		  ( 
		    proc_size ==2 or (debug_proc BASE_PROC "MULTI: Sent a msg to INV GEN"; isend STOP 2 9 comm_world;true);
		    local_base_stop := true )
	      end
      | STOP ->
	  begin
	    Globals.base_time_stop := (wtime());
	    let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
	    let bt_string = string_of_float base_time in 
	      if !Flags.xml then (
		print_xml (Kind_support.print_resultProp_xml 
			     {result=UNKNOWN;
			      foundvars=None;
			      minstep=None;
			      maxstep=None;
			      induct_cex= None;
			      name="UNKNOWN RESULT";
			      k="";
			      time=bt_string});
	      ) else (
		print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
		print_to_user_mpkind ("  UNKNOWN RESULT (maximum K iterations reached)  \n");
		print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n"););
	      
	      if !Flags.mpi_abort then (
		if !Flags.xml then (
		  print_xml ("</Results>");
		);
		mpi_abort()
	      ) else ( 
		proc_size ==2 or (isend STOP 2 9 comm_world;true);
		Globals.base_stop := true )
	  end
      |  _ -> 
	   begin
	     print_to_user_mpkind ("MULTI: Received wrong message.\n");
	     Globals.base_time_stop := (wtime());
	     local_base_stop := true
	   end
end;   

(*Check timeout*)
    if ((wtime()) -. !Globals.base_time_start) >= !Flags.timeout then (
      Globals.base_time_stop := (wtime());
      let base_time = !Globals.base_time_stop -. !Globals.base_time_start in
      let bt_string = string_of_float base_time in 
	if !Flags.xml then (
	  let prop_names =  Multi_prop_util.mk_prop_names !unsolved_prop_list in
	  print_xml (Kind_support.print_resultProp_xml 
		       {result=TIMEOUT;
			foundvars=None;
			minstep=None;
			maxstep=None;
			induct_cex= None;
			name=prop_names;
			k="";
			time=bt_string})
	) else(
	  print_to_user_mpkind("\n\n++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
	  print_to_user_mpkind ("                       TIMEOUT                           \n");
	  print_to_user_mpkind("++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n"););
	if !Flags.mpi_abort then
	  (mpi_abort();
	  )else(
	    let _ = isend STOP step_proc 10 comm_world in
	    let _ = proc_size ==2 or ( debug_proc BASE_PROC "Sent a msg to INV GEN";isend STOP 2 9 comm_world;true) in
	      local_base_stop := true));
 done (* main loop *) 


	






